local Mod = LibStub("AceAddon-3.0"):NewAddon("Broker Hittable", "AceEvent-3.0", "AceBucket-3.0")

Broker_Hittable = Mod

local L = LibStub("AceLocale-3.0"):GetLocale("Broker Hittable", false)

local LibQTip = LibStub('LibQTip-1.0')

local OUTCOMES = {
	"MISS",
	"DODGE",
	"PARRY",
	"BLOCK",
	"CRIT",
	"CRUSHING",
	"HIT",
}

local COLORS = {
	["MISS"] = "|cff69b950",
	["DODGE"] = "|cff69b950",
	["PARRY"] = "|cff69b950",
	["BLOCK"] = "|cff00ff00",
	["CRIT"] = "|cffff0000",
	["CRUSHING"] = "|cffff7f00",
	["HIT"] = "|cffffff00",
}

local SPELL_DODGE = GetSpellInfo(81)
local SPELL_PARRY = GetSpellInfo(3127)
local SPELL_BLOCK = GetSpellInfo(107)

local ICON_DW_DISABLED = "Interface\\Icons\\Ability_MeleeDamage"
local ICON_DW_ENABLED = "Interface\\Icons\\Ability_DualWield"

local defaults = {
	char = {
		enabled = true,
		dualwield = false
	},
	profile = {
		throttleinterval = 1.5,
		defaultlevel = 83,
		usetarget = true,
		bosslevel = 83
	}
}

local db
local options

local LDBObject
local tooltip = nil
local auraBucket

local hitTable = {}
local mobLevel
local defaultLevel
local bossLevel

-- Upvalues
local IsUsableSpell, GetDodgeChance, GetParryChance, GetBlockChance, GetCombatRatingBonus = IsUsableSpell, GetDodgeChance, GetParryChance, GetBlockChance, GetCombatRatingBonus
local UnitDefense, UnitExists, UnitIsEnemy, UnitLevel, UnitClassification, UnitClass = UnitDefense, UnitExists, UnitIsEnemy, UnitLevel, UnitClassification, UnitClass

local function FillTooltip()
	tooltip:Clear()
	local y, x = tooltip:AddLine()
	tooltip:SetCell(y, x, "|cffffd200"..L["TITLE"], "CENTER", 2)
	tooltip:AddLine(" ", " ")
	if Mod:IsEnabled() then
		tooltip:AddLine("|cffffff9a"..L["OUTCOME"], "|cffffff9a"..L["CHANCE"])
		for _, v in pairs(OUTCOMES) do
			if hitTable[v] > 0.0 then
				tooltip:AddLine(COLORS[v]..L[v], string.format("%.2f", hitTable[v]))
			end
		end

		tooltip:AddLine(" ", " ")
		y = tooltip:AddLine()
		tooltip:SetCell(y, 1, "|cffffff9a"..string.format(L["LEVELTEXT"], mobLevel), 2)
		if(db.char.dualwield) then
			y = tooltip:AddLine()
			tooltip:SetCell(y, 1, "|cffffff9a"..L["DWTEXT"], 2)
		end
		tooltip:AddLine(" ", " ")
		y, x = tooltip:AddLine()
		tooltip:SetCell(y, x, "|cffffd200"..L["ENABLED_NOTE"], GameTooltipTextSmall, 2)
	else
		y, x = tooltip:AddLine()
		tooltip:SetCell(y, x, "|cffffd200"..L["DISABLED_NOTE"], GameTooltipTextSmall, 2)
	end
end

local function UpdateTooltip()
	local lastTableLine = tooltip:GetLineCount() - (4 + (db.char.dualwield and 1 or 0))
	local line = 4
	local font = tooltip:GetFont()
	for _, v in pairs(OUTCOMES) do
		if hitTable[v] > 0.0 then
			if(line > lastTableLine) then
				tooltip:AddLine(" ")
			end
			tooltip:SetCell(line, 1, COLORS[v]..L[v], font, 1)
			tooltip:SetCell(line, 2, string.format("%.2f", hitTable[v]), font, 1)
			line = line + 1
		end
	end
	if line <= lastTableLine then
		repeat
			tooltip:SetCell(line, 1, "", font, 1)
			tooltip:SetCell(line, 2, "", font, 1)
			line = line + 1
		until line > lastTableLine
	elseif line > lastTableLine + 1 then
		tooltip:SetCell(line, 1, "", font, 1)
		tooltip:SetCell(line, 2, "", font, 1)
		tooltip:SetCell(line+1, 1, "|cffffff9a"..string.format(L["LEVELTEXT"], mobLevel), font, 2)
		if(db.char.dualwield) then
			tooltip:SetCell(line+2, 1, "|cffffff9a"..L["DWTEXT"], font, 2)
			line = line + 1
		end
		tooltip:SetCell(line+2, 1, "", font, 1)
		tooltip:SetCell(line+2, 2, "", font, 1)
		tooltip:SetCell(line+3, 1, "|cffffd200"..L["ENABLED_NOTE"], GameTooltipTextSmall, 2)
	end
end

local ToggleDualWield, InitializeHittable, OnInventoryOrAurasChanged, OnBaseDefenseChanged, OnTalentsChanged
local UpdateHittable = function() end -- dummy function, gets replaced after InitializeHittable
local SetMobLevel = function(level) mobLevel = level end -- another dummy
do
	local baseMiss

	ToggleDualWield = function()
		if db.char.dualwield then
			db.char.dualwield = false
			baseMiss = baseMiss - 19
			LDBObject.icon = ICON_DW_DISABLED
		else
			db.char.dualwield = true
			baseMiss = baseMiss + 19
			LDBObject.icon = ICON_DW_ENABLED
		end
		UpdateHittable(false)
		if(tooltip) then
			FillTooltip()
		end
	end

	local RealUpdateHittable
	local k
	local class
	local CLASS_CONSTANTS = {
		["DEATHKNIGHT"] = 0.956,
		["DRUID"] = 0.972,
		["HUNTER"] = 0.988,
		["MAGE"] = 0.953,
		["PALADIN"] = 0.956,
		["PRIEST"] = 0.953,
		["ROGUE"] = 0.988,
		["SHAMAN"] = 0.988,
		["WARLOCK"] = 0.953,
		["WARRIOR"] = 0.956,
	}

	InitializeHittable = function()
		_, class = UnitClass("player")
		k = CLASS_CONSTANTS[class]
		baseMiss = 5
		if(select(2, UnitRace("player")) == "NightElf") then
			 baseMiss = baseMiss + 2
		end
		if(db.char.dualwield) then
			baseMiss = baseMiss + 19
		end
		
		UpdateHittable = function() end
		
		OnBaseDefenseChanged()
		OnTalentsChanged()
		OnInventoryOrAurasChanged({["player"] = 1})
		
		UpdateHittable = RealUpdateHittable
		SetMobLevel = RealSetMobLevel

		UpdateHittable(false)
	end

	local baseDefense
	local bonusDefense

	OnInventoryOrAurasChanged = function(unitIds)
		if(unitIds["player"]) then
			_, bonusDefense = UnitDefense("player")
			UpdateHittable(tooltip)
		end
	end

	local critReductionFromTalents

	OnTalentsChanged = function()
		critReductionFromTalents = 0
		if class == "DRUID" then
			critReductionFromTalents = select(5, GetTalentInfo(2, 18))*2
		end
		UpdateHittable(tooltip)
	end

	local skillDifference
	local chanceFromSkill

	OnBaseDefenseChanged = function()
		baseDefense = UnitDefense("player")
		skillDifference = mobLevel*5 - baseDefense
		chanceFromSkill = 0.04*skillDifference
		UpdateHittable(tooltip)
	end

	RealSetMobLevel = function(level)
		if(level ~= mobLevel) then
			skillDifference = level*5 - baseDefense
			chanceFromSkill = 0.04*skillDifference
			mobLevel = level
			UpdateHittable(tooltip)
		end
	end

	local totalChance
	local avoidance

	local function SetHittableValue(outcome, probability)
		v = max(0, min(100-totalChance, probability))
		totalChance = totalChance + v
		hitTable[outcome] = v
	end

	RealUpdateHittable = function(updatetooltip)
		totalChance = 0.0

		SetHittableValue("MISS", baseMiss + 1/(0.0625 + k/(0.04*bonusDefense)) - chanceFromSkill)
		SetHittableValue("DODGE", (IsUsableSpell(SPELL_DODGE) and GetDodgeChance()-chanceFromSkill) or 0)
		SetHittableValue("PARRY", (IsUsableSpell(SPELL_PARRY) and GetParryChance()-chanceFromSkill) or 0)
		avoidance = totalChance
		SetHittableValue("BLOCK", (IsUsableSpell(SPELL_BLOCK) and GetBlockChance()-chanceFromSkill) or 0)
		SetHittableValue("CRIT", 5 - 0.04*bonusDefense - GetCombatRatingBonus(CR_CRIT_TAKEN_MELEE) - critReductionFromTalents + chanceFromSkill)
		SetHittableValue("CRUSHING", ((skillDifference > 15) and 2*skillDifference-15) or 0)
		hitTable["HIT"] = 100 - totalChance

		if(updatetooltip) then
			UpdateTooltip()
		end
		LDBObject.text = string.format("%.2f%%", avoidance)
	end
end

local function OnTargetChanged()
	if UnitIsEnemy("player", "target") then
		local level = UnitLevel("target")
		if level == -1 then
			if UnitClassification("target") == "worldboss" then
				level = bossLevel
			else
				level = UnitLevel("player") + 10
			end
		end
		SetMobLevel(level)
	else
		SetMobLevel(defaultLevel)
	end
end

local function OnEnter(button)
	tooltip = LibQTip:Acquire("Broker Hittable", 2, "LEFT", "RIGHT")
	FillTooltip()
	tooltip:SmartAnchorTo(button)
	tooltip:Show()
end

function OnLeave()
	LibQTip:Release(tooltip)
	tooltip = nil
end

local function OnClick(frame, button)
	if Mod:IsEnabled() then
		if button == "RightButton" then
			Mod:Disable()
		else
			ToggleDualWield()
		end
	else
		Mod:Enable()
	end
end

local function RefreshConfig()
	defaultLevel = db.profile.defaultlevel
	bossLevel = db.profile.bosslevel

	if(db.char.enabled) then -- This is the not the initial setup
		Mod:UnregisterBucket(auraBucket)
		-- Update table in case we missed events from the deleted bucket
		OnInventoryOrAurasChanged({["player"] = 1})
		Mod:UnregisterEvent("PLAYER_TARGET_CHANGED")
	end

	auraBucket = Mod:RegisterBucketEvent({"UNIT_AURA", "UNIT_INVENTORY_CHANGED"}, db.profile.throttleinterval, OnInventoryOrAurasChanged)

	if(db.profile.usetarget) then
		Mod:RegisterEvent("PLAYER_TARGET_CHANGED", OnTargetChanged)
		OnTargetChanged() -- Force update
	else
		SetMobLevel(defaultLevel)
	end
end

function Mod:OnInitialize()
	db = LibStub("AceDB-3.0"):New("Broker_HittableDB", defaults)
	options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)
	options.args.profiles.dialogHidden = true
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Broker_Hittable", options)
	local ACD = LibStub("AceConfigDialog-3.0")
	ACD:AddToBlizOptions("Broker_Hittable", L["TITLE"])
	ACD:AddToBlizOptions("Broker_Hittable", L["PROFILES"], L["TITLE"], "profiles")
	LibStub("LibAboutPanel").new(L["TITLE"], "Broker_Hittable")

	self:SetEnabledState(db.char.enabled) -- Stop auto-enabling if addon is disabled

	LDBObject = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject(L["TITLE"], {
		type = "data source",
		label = L["LABEL"],
		text = "",
		icon = db.char.dualwield and ICON_DW_ENABLED or ICON_DW_DISABLED,
		OnClick = OnClick,
		OnEnter = OnEnter,
		OnLeave = OnLeave,
	})
end

function Mod:OnEnable()
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", OnTalentsChanged)
	self:RegisterEvent("PLAYER_TALENT_UPDATE", OnTalentsChanged) -- New in 3.1: Fires when switching specs
	self:RegisterEvent("CHAT_MSG_SKILL", OnBaseDefenseChanged)
	db.RegisterCallback(self, "OnProfileChanged", RefreshConfig)
	db.RegisterCallback(self, "OnProfileCopied", RefreshConfig)
	db.RegisterCallback(self, "OnProfileReset", RefreshConfig)

	RefreshConfig()
	InitializeHittable()

	if(tooltip) then
		FillTooltip()
	end

	db.char.enabled = true
end

function Mod:OnDisable()
	self:UnregisterBucket(auraBucket)
	self:UnregisterAllEvents()
	db.UnregisterAllCallbacks(self, false)

	db.char.enabled = false

	if(tooltip) then
		FillTooltip()
	end
	LDBObject.text = ""
end

options = {
   type = "group",
   get = function(info) return db[info.option.arg][info[1]] end,
   args = {
      enabled = {
         type = "toggle",
         name = L["OPT_ENABLED"],
         desc = L["OPT_ENABLED_DESC"],
         arg = "char",
         set = function(_, value) if(value) then Mod:Enable() else Mod:Disable() end end,
         order = 0,
      },
      dualwield = {
         type = "toggle",
         name = L["OPT_DUALWIELD"],
         desc = L["OPT_DUALWIELD_DESC"],
         arg = "char",
         set = function(_, value) ToggleDualWield() end,
         order = 1,
      },
      throttleinterval = {
         type = "range",
         min = 0.1,
         max = 5.0,
         step = 0.35,
         name = L["OPT_THROTTLE"],
         desc = L["OPT_THROTTLE_DESC"],
         arg = "profile",
         set = function(_, value)
		Mod:UnregisterBucket(auraBucket)
		OnInventoryOrAurasChanged({["player"] = 1})
		auraBucket = Mod:RegisterBucketEvent({"UNIT_AURA", "UNIT_INVENTORY_CHANGED"}, value, OnInventoryOrAurasChanged)
		db.profile.throttleinterval = value
         end,
         order = 2,
      },
      defaultlevel = {
         type = "range",
         min = 1,
         max = 83,
         step = 1,
         name = L["OPT_DEFAULTLEVEL"],
         desc = L["OPT_DEFAULTLEVEL_DESC"],
         arg = "profile",
         set = function(_, value)
         	defaultLevel = value
         	db.profile.defaultlevel = value
         	if(db.profile.usetarget) then
         		OnTargetChanged()
         	else
         		SetMobLevel(defaultLevel)
         	end
         end,
         order = 3,
      },
      usetarget = {
         type = "toggle",
         name = L["OPT_USETARGET"],
         desc = L["OPT_USETARGET_DESC"],
         arg = "profile",
         set = function(_, value)
         	db.profile.usetarget = value
         	if(value) then
			Mod:RegisterEvent("PLAYER_TARGET_CHANGED", OnTargetChanged)
			OnTargetChanged() -- Force update
		else
       			Mod:UnregisterEvent("PLAYER_TARGET_CHANGED")
       			SetMobLevel(defaultLevel)
       		end
         end,
         order = 4,
      },
      bosslevel = {
         type = "range",
         min = 1,
         max = 83,
         step = 1,
         name = L["OPT_BOSSLEVEL"],
         desc = L["OPT_BOSSLEVEL_DESC"],
         arg = "profile",
         set = function(_, value)
         	bossLevel = value
         	db.profile.bosslevel = value
       		OnTargetChanged()
         end,
         disabled = function() return not db.profile.usetarget end,
         order = 5,
      },
   }
}
